`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: Steven Howell
// 
// Create Date:    09:10:13 03/12/2015 
// Design Name: 
// Module Name:    i2c_rom 
// Project Name: 
// Target Devices: XC7A35TCSG324-1L
// Tool versions: Vivado 2017.1
// Description: initialization contents for I2C writes to HMC5883L
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////



module i2c_rom #(parameter HMC5883L_ADDR = 8'h3C, CONFIG_A = 8'd00, CONFIG_B = 8'd01, MODE_REG = 8'd02, 
                 DATA_X_MSB = 8'd03, DATA_X_LSB = 8'd04, DATA_Z_MSB = 8'd05, DATA_Z_LSB = 8'd06, DATA_Y_MSB = 8'd07, DATA_Y_LSB = 8'd08, 
                 STATUS = 8'd09, ID_A = 8'd10, ID_B = 8'd11, ID_C = 8'd12)
						
					(input [7:0] addr, output reg [7:0] slave_addr, output reg [7:0] reg_addr, output reg [7:0] data, output reg WR, output reg RD, output reg retry, output reg [2:0] rd_bytes,
					output reg no_data);
	 
	 always @(addr)
        case (addr)
		  
//                                                                             					I2C slave addr     		byte addr        data (if req'd)   write   read  retry     # of read
//                                                                               					(8 bits)         		(8 bits)           (8 bits)                  (y/n)      bytes    	no data				  
              // set up the read pointer, configuration and mode registers			  						
			  8'h00		:		{slave_addr, reg_addr, data, WR, RD, retry, rd_bytes, no_data} <= {HMC5883L_ADDR,		8'h00,     				8'h00, 		1'b1, 	1'b0, 1'b0, 	3'b000, 	1'b1};					// set register pointer to 0, write, no data word																						  
		     
		      // configuration register A, CRA - 7 cleared, CRA 6:5 - number of samples averaged per output, CRA4:2 - data output rate selection, CRA1:0 - measurement bias used or not
		      8'h01		:		{slave_addr, reg_addr, data, WR, RD, retry, rd_bytes, no_data} <= {HMC5883L_ADDR,		CONFIG_A,           8'b01111000, 	1'b1, 	1'b0, 1'b0, 	3'b000,		1'b0};		
			  8'h02		:		{slave_addr, reg_addr, data, WR, RD, retry, rd_bytes, no_data} <= {HMC5883L_ADDR,		CONFIG_B,        	8'b00000000, 	1'b1, 	1'b0, 1'b0, 	3'b000,		1'b0};					//
			  8'h03		:		{slave_addr, reg_addr, data, WR, RD, retry, rd_bytes, no_data} <= {HMC5883L_ADDR,		MODE_REG,           8'b10000000, 	1'b1, 	1'b0, 1'b0, 	3'b000,		1'b0};					//
			  
			  // read back the status register
			  8'h04		:		{slave_addr, reg_addr, data, WR, RD, retry, rd_bytes, no_data} <= {HMC5883L_ADDR,		STATUS,                 8'h00, 		1'b0, 	1'b1, 1'b0, 	3'b011, 	1'b0};

			  // read the X, Y, Z values
			  8'h05		:		{slave_addr, reg_addr, data, WR, RD, retry, rd_bytes, no_data} <= {HMC5883L_ADDR,		DATA_X_MSB,				8'h00, 		1'b0, 	1'b1, 1'b0, 	3'b101, 	1'b0};

			  
          default		:  	begin
							slave_addr <= 8'h00;
							reg_addr <= 8'h00;
							data <= 8'h00;		
							WR <= 1'b0;
							RD <= 1'b0;
							retry <= 1'b0;
							rd_bytes <= 3'b000;
							no_data <= 1'b0;
							end
         endcase



endmodule
