`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: Steven Howell
// 
// Create Date: 07/17/2017 06:41:35 PM
// Design Name: 
// Module Name: top_hdg_ctrl
// Project Name: 
// Target Devices: XC7A35TCSG324-1L
// Tool Versions: Vivado 2017.1
// Description: motor control logic for Whippersnapper Runt
// 
// Dependencies: 
// 
// Revision:
// Revision 0.01 - File Created
// Additional Comments:
// 
//////////////////////////////////////////////////////////////////////////////////


module top_hdg_ctrl (input clk, input cal_complete, output motor_enable,
                    output reg A_1A, output reg A_1B, output reg A_2A, output reg A_2B, // left side
                    output reg B_1A, output reg B_1B, output reg B_2A, output reg B_2B, // right side
                    input [23:0] current_hdg, input [23:0] cmd_hdg, input fwd, input rev);
   
localparam PI_4     =   24'b0000_0000_0011_0010_0100_0010;  // .79 in real, fractional 14 bits   
localparam PI_2     =   24'b0000_0000_0110_0100_0100_0000;  // 1.57 in real, fractional 14 bits
localparam tPI_4    =   24'b0000_0000_1001_0110_1100_0111;  // 2.36 in real, fractional 14 bits
localparam PI       =   24'b0000_0000_1100_1000_1100_0000;  // 3.14 in real, fractional 14 bits
localparam tPI_2    =   24'b0000_0001_0010_1101_0100_0000;  // 4.71 in real, fractional 14 bits
localparam tPI      =   24'b0000_0001_1001_0001_1100_0000;  // 6.28 in real, fractional 14 bits 

localparam ERR_THOLD=   24'b0000_0000_0001_0000_0000_0000;  // .25 radians is where the erorr will prompt motor activity--fractional 14 bits
//localparam ERR_THOLD=   24'b0000_0000_0000_1000_0000_0000;  // .125 radians is where the erorr will prompt motor activity--fractional 14 bits

localparam FIRST_GEAR   =   12'h880;
localparam SECOND_GEAR  =   12'h900;
localparam THIRD_GEAR   =   12'h980;
localparam FOURTH_GEAR  =   12'hA80;            // this value cannot be exceeded if using the xtra!!
                                      
wire [23:0] left_error, right_error, error;
reg [11:0] PWM_counter, speed;
reg slow, drive, xtra;
  

// A LOOP
// The "A Loop" below sends out the signals going to the tristate enable ports of the buffers at the top level
always@(posedge clk)
begin
    if (motor_enable)
    begin
        if (fwd)           // go forward and correct heading error simultaneously
            begin
                    
            if (error == right_error)
                begin
                A_1A <= xtra;                       // left front motor fwd xtra            
                A_1B <= 1'b0;
                A_2A <= xtra;                       // left rear motor fwd plus extra
                A_2B <= 1'b0;
                B_1A <= slow;                      // both right side motors ahead drive
                B_1B <= 1'b0;
                B_2A <= slow;
                B_2B <= 1'b0;
                end
                else 
                    begin
                    A_1A <= slow;                  // both left side motors ahead drive        
                    A_1B <= 1'b0;
                    A_2A <= slow;             
                    A_2B <= 1'b0;
                    B_1A <= xtra;                   // right side motors fwd xtra 
                    B_1B <= 1'b0;
                    B_2A <= xtra;                   
                    B_2B <= 1'b0;          
                    end

            end
            else if (rev)
                begin
                
                if (error == right_error)
                    begin
                    A_1A <= 1'b0;                       // both left side motors back drive            
                    A_1B <= drive;
                    A_2A <= 1'b0;                       
                    A_2B <= drive;
                    B_1A <= 1'b0;                      
                    B_1B <= xtra;                       // right side motors back extra
                    B_2A <= 1'b0;
                    B_2B <= xtra;                        
                    end
                    else 
                        begin
                        A_1A <= 1'b0;                  // both left side motors back extra        
                        A_1B <= xtra;
                        A_2A <= 1'b0;             
                        A_2B <= xtra;
                        B_1A <= 1'b0;                   // right side motors back drive 
                        B_1B <= drive;
                        B_2A <= 1'b0;                   
                        B_2B <= drive;          
                        end
                
                
                end
                else
                    begin
                    
                    if (error == right_error)
                        begin
                        // no forward or reverse command, but need to correct heading to the right
                        A_2A <= xtra;                  // left motors fwd
                        A_2B <= 1'b0;
                        A_1A <= xtra;
                        A_1B <= 1'b0;
                                            
                        B_2B <= 1'b0;                  // right motors drive ahead
                        B_2A <= slow;
                        B_1B <= 1'b0;                  
                        B_1A <= slow;
                        end
                        else
                            begin
                            // no forward or reverse command, but need to correct heading to the left
                            B_2A <= xtra;              // right motors fwd
                            B_2B <= 1'b0;
                            B_1A <= xtra;                   
                            B_1B <= 1'b0;
                                                
                            A_2B <= 1'b0;              // left motors drive ahead
                            A_2A <= slow;
                            A_1B <= 1'b0;                   
                            A_1A <= slow;
                            end

                    end
        
    end
    else
        begin
         A_1A <= 1'b0;                  // if there is no motor enable, then no motors should drive       
         A_1B <= 1'b0;
         A_2A <= 1'b0;             
         A_2B <= 1'b0;
         B_1A <= 1'b0;                   
         B_1B <= 1'b0;
         B_2A <= 1'b0;                   
         B_2B <= 1'b0;          
        end

end 
// END A LOOP

// B LOOP
// The "B loop" assigns the speed of wheel motors based on the difference in heading.  The value is placed into the register "speed" which is used in "loop D."
always@(posedge clk)
begin
    if (error >= tPI_4) speed <= FOURTH_GEAR;
        else if (error >= PI_2) speed <= THIRD_GEAR;
            else if (error >= PI_4) speed <= SECOND_GEAR;
                else speed <= FIRST_GEAR;
end
// END B LOOP


// C LOOP
assign left_error = (cmd_hdg >= current_hdg) ? cmd_hdg - current_hdg : (cmd_hdg + tPI) - current_hdg;       // left error always positive, when greater than PI, should turn right
assign right_error = (cmd_hdg >= current_hdg) ? (tPI - cmd_hdg) + current_hdg : current_hdg - cmd_hdg;      // right error always positive, when greather than PI, should turn left 
assign error = (left_error < right_error) ? left_error : right_error; 

// if the commanded heading is more than ~14 degrees, or fwd/rev is commanded, and the mag calibration is complete, enable the motors
assign motor_enable = (((error >= ERR_THOLD) | (fwd) | (rev)) && cal_complete) ? 1'b1 : 1'b0;
// END C LOOP
   

// D LOOP
// start the PWM counter, and increment it on every clock.  Drive the signals "drive" and "xtra" depending on the value of the count.
// "drive" is the pulse width that will be applied to wheels that are turning at the normal rate for a particular speed.  "xtra" will be applied to 
// turn some wheels slightly faster to cause a turn when the vehicle is already going forward or reverse at the normal "drive" speed                    
always@(posedge clk)
begin
    if (motor_enable)                                       // if the motor is enabled
        begin
        PWM_counter <= PWM_counter + 1'b1;                  // increment the counter for PWM
        
        if (PWM_counter < (speed>>2)) slow <= 1'b1;         // slow is a barely roll forward value
            else slow <= 1'b0;
        
        if (PWM_counter < speed) drive <= 1'b1;             // drive used for correction of heading
            else drive <= 1'b0;
            
        if (PWM_counter < (speed + (speed>>2))) xtra <= 1'b1;   // xtra used for forward or reverse
            else xtra <= 1'b0;

        end
        else
            begin
            PWM_counter <= 12'h000;
            slow <= 1'b0;
            drive <= 1'b0;
            xtra <= 1'b0;
            end

end 
// END D LOOP                   
                    
endmodule
